import 'dart:async';
import 'package:adeptannotations/adeptannotations.dart';
import 'package:adeptgenerator/src/base_generator_for_annotation.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/src/builder/build_step.dart';
import 'package:source_gen/source_gen.dart';

class DiffGenerator
    extends BaseGeneratorForAnnotation<Differentiable, DiffKey> {
  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    ClassElement classElement =
        getClassElement(element, annotationName: "@Differentiable()");
    String diffFunction = _generateDiffFunction(classElement, annotation);
    return '''
    $diffFunction
    ''';
  }

  /// This function manages the logic of generating the Diff boilerplate code
  String _generateDiffFunction(
      ClassElement classElement, ConstantReader annotation) {
    checkDefaultConstructor(classElement, annotationName: "@Differentiable()");
    String classType = classElement?.displayName;
    String functionName = "\$${classType}Diff";
    String currentDataParam = "currentData";
    String newDataParam = "newData";
    String functionParams = "{$classType $currentDataParam, "
        "$classType $newDataParam}";
    String diffVar = "diff";
    String isDiffVar = "isDiff";

    StringBuffer diffData = StringBuffer("final $diffVar = $classType()");
    Differentiable differentiable =
        _getDifferentiable(classElement) ?? Differentiable();
    classElement?.fields?.forEach((field) {
      if (field.isSynthetic && differentiable.ignoreSynthetic) return;
      if (field.isFinal || field.isConst || field.isStatic) return;
      if (field.isSynthetic && field.getter == null) return;
      DiffKey diffKey = _getDiffKey(field);
      if (diffKey?.ignore ?? false) return;
      String fieldName = field?.displayName;
      String diffDataFieldValue = "$newDataParam?.$fieldName != "
          "$currentDataParam?.$fieldName ? $newDataParam?.$fieldName : null";
      if (diffKey?.forceCurrentData ?? false)
        diffDataFieldValue = "$currentDataParam?.$fieldName";
      else if (diffKey?.forceNewData ?? false)
        diffDataFieldValue = "$newDataParam?.$fieldName";
      diffData.writeln("..${fieldName} = $diffDataFieldValue");
    });
    diffData.writeln(";");

    StringBuffer  isDiffData = StringBuffer("");
    classElement?.fields?.forEach((field) {
      if (field.isSynthetic && differentiable.ignoreSynthetic) return;
      if (field.isFinal || field.isConst || field.isStatic) return;
      if (field.isSynthetic && field.getter == null) return;
      String fieldName = field?.displayName;
      isDiffData.writeln("$isDiffVar |= $diffVar?.$fieldName != null;");
    });

    return '''
      /// This function creates a new instance of [$classType] with only the 
      /// fields that are not equal (==) between [$currentDataParam] and 
      /// [$newDataParam] 
      $classType $functionName($functionParams){
        
        if(newData == null) return currentData;
        if(currentData == null) return newData;
        if(newData == null && currentData == null) return null;
        
        ${diffData.toString()}
        
        bool $isDiffVar = false;
        ${isDiffData.toString()}
        
        return $isDiffVar ? $diffVar : null;
      }
    ''';
  }

  /// Gets the Differentiable annotation for a given ClassElement
  Differentiable _getDifferentiable(ClassElement classElement) {
    DartObject differentiableAnnotation = annotationChecker
        .firstAnnotationOf(classElement, throwOnUnresolved: false);
    if (differentiableAnnotation != null)
      return Differentiable(
          ignoreSynthetic: differentiableAnnotation
              ?.getField('ignoreSynthetic')
              ?.toBoolValue());
    return null;
  }

  /// Gets the DiffKey annotation for a given FieldElement
  DiffKey _getDiffKey(FieldElement field) {
    Iterable<DartObject> annotations =
        keyAnnotationChecker.annotationsOfExact(field);
    int annotationsCount = annotations?.length ?? 0;
    if (annotationsCount > 1)
      throw InvalidGenerationSourceError(
          'Repeated `@DiffKey` annotation on ${field?.displayName}.',
          todo: 'Check the `@DiffKey` annotations on field '
              '${field?.displayName}.',
          element: field);
    if (annotationsCount == 1) {
      DartObject annotation = annotations.first;
      final key = DiffKey(
        ignore: annotation.getField('ignore')?.toBoolValue(),
        forceCurrentData:
            annotation.getField('forceCurrentData')?.toBoolValue(),
        forceNewData: annotation.getField('forceNewData')?.toBoolValue(),
      );
      if ((key.forceCurrentData ?? false) && (key.forceNewData ?? false))
        throw InvalidGenerationSourceError(
            'forceCurrentData and forceNewData properties on '
            '${field?.displayName} are mutually exclusive and can\'t be true '
            'on the same field.',
            todo: 'Check the `@DiffKey` properties on field '
                '${field?.displayName}.',
            element: field);
      return key;
    }
    return null;
  }
}
